"""Landmark-based Morphometrics"""

# Copyright 2020 Koji Noshita
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sklearn.base import BaseEstimator, TransformerMixin


class PositionAligner(TransformerMixin, BaseEstimator):
	"""Align the coordinate values .
	This estimator scales and translates each feature individually such
	that it is in the given range on the training set, e.g. between
	zero and one.
	The transformation is given by::
	    X_std = (X - X.min(axis=0)) / (X.max(axis=0) - X.min(axis=0))
	    X_scaled = X_std * (max - min) + min
	where min, max = feature_range.
	This transformation is often used as an alternative to zero mean,
	unit variance scaling.
	Read more in the :ref:`User Guide <preprocessing_scaler>`.
	Parameters
	----------
	copy : bool, default=True
	    Set to False to perform inplace row normalization and avoid a
	    copy (if the input is already a numpy array).
	Attributes
	----------
	data_range_ : ndarray of shape (n_features,)
	    Per feature range ``(data_max_ - data_min_)`` seen in the data
	    .. versionadded:: 0.17
	       *data_range_*
	n_samples_seen_ : int
	    The number of samples processed by the estimator.
	    It will be reset on new calls to fit, but increments across
	    ``partial_fit`` calls.
	Examples
	--------
	>>> from sklearn.preprocessing import MinMaxScaler
	>>> data = [[-1, 2], [-0.5, 6], [0, 10], [1, 18]]
	>>> scaler = MinMaxScaler()
	>>> print(scaler.fit(data))
	MinMaxScaler()
	>>> print(scaler.data_max_)
	[ 1. 18.]
	>>> print(scaler.transform(data))
	[[0.   0.  ]
	 [0.25 0.25]
	 [0.5  0.5 ]
	 [1.   1.  ]]
	>>> print(scaler.transform([[2, 2]]))
	[[1.5 0. ]]
	See Also
	--------
	minmax_scale : Equivalent function without the estimator API.
	Notes
	-----
	NaNs are treated as missing values: disregarded in fit, and maintained in
	transform.
	For a comparison of the different scalers, transformers, and normalizers,
	see :ref:`examples/preprocessing/plot_all_scaling.py
	<sphx_glr_auto_examples_preprocessing_plot_all_scaling.py>`.
	"""
	def __init__(self, copy=True):
		self.copy = copy

	def transform(self, X, reference_point=None):
	"""Perform alignment by centroid position
	Parameters
	----------
	X : {array-like, dense matrix of shape (n_samples, n_landmarks, n_dims)
	    The coordinate values to be aligned
	reference_point: {array-like}
 		The coordinate value of the reference point to be aligned onto the origin. 
 		If `reference_point` is `None`, the centroid computed from `X` is used as the reference point. 
	Returns
	-------
	X_tr : {ndarray, sparse matrix} of shape (n_samples, n_landmarks,n_dims)
	    Transformed array, which is removed the information about the position.
	"""
	check_is_fitted(self)

	n_samples, _, n_dims = X.shape
	X_tr = X - X.mean(axis=1).reshape(n_samples,1,n_dims)

	return X_tr

class SizeScaler(TransformerMixin, BaseEstimator):
	def __init__(self):
		self.copy = copy

	def transform(self, X):
	"""Perform scaling by centroid size
	Parameters
	----------
	X : {array-like, dense matrix of shape (n_samples, n_landmarks, n_dims)
	    The coordinate values used to align by the centroid position onto the origin.
	Returns
	-------
	X_tr : {ndarray, sparse matrix} of shape (n_samples, n_landmarks,n_dims)
	    Transformed array, which is removed the information about the position.
	"""


class OrientationAligner(TransformerMixin, BaseEstimator):
	def __init__(self):


class ProcrustesSuperImposer(TransformerMixin, BaseEstimator):
	def __init__(self):




import pytest

from sklearn.utils.estimator_checks import check_estimator

from skltemplate import TemplateEstimator
from skltemplate import TemplateClassifier
from skltemplate import TemplateTransformer


@pytest.mark.parametrize(
    "Estimator", [TemplateEstimator, TemplateTransformer, TemplateClassifier]
)
def test_all_estimators(Estimator):
    return check_estimator(Estimator)



"""Spherical Harmonic (SPHARM) Analysis """

# Copyright 2020 Koji Noshita
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import scipy as sp

from sklearn.base import BaseEstimator, TransformerMixin


class SphericalHarmonicAnalysis(TransformerMixin, BaseEstimator):
    r"""Spherical Harmonic (SPHARM) Analysis


    Notes
    ------------------------
    [Ritche_Kemp_1999]_, [Shen_etal_2009]_

    .. math::
        \begin{align}
            \mathbf{p}(\theta, \phi) = \sum_{l=0}^{l_\mathrm{max}} \sum_{m=-l}^l
            \left(
                c_l^m Y_l^m(\theta, \phi)
            \right)
        \end{align}

    , where :math:`Y_l^m(\theta, \phi)` are spherical harmonics:

    .. math::
        \begin{align}
            Y_l^m(\theta, \phi) = \sqrt{\frac{2l+1}{4\pi}\frac{(l-m)!}{(l+m)!}} P_l^m(\cos(\theta)) e^{im\phi}
        \end{align}

    , where :math:`P_n^m(x)` are associated Legendre polynomials:

    .. math::

        \begin{align}
            P_n^m(x) = (-1)^m (1-x^2)^{\frac{m}{2}} \frac{d^m}{dx^m} P_n(x)
        \end{align}

    , where :math:`P_n(x)` are Legendre polynomials, which are solutions of Legendre’s differential equation;

    .. math::
        (1-x^2)\frac{d^2 y}{dx^2} -2x \frac{dy}{dx} + n(n+1)y = 0.


    References
    ------------------------

    .. [Ritche_Kemp_1999] Ritchie, D.W., Kemp, G.J.L. (1999) Fast computation, rotation, and comparison of low resolution spherical harmonic molecular surfaces. J. Comput. Chem. 20: 383–395.
    .. [Shen_etal_2009] Shen, L., Farid, H., McPeek, M.A. (2009) Modeling three-dimensional morphological structures using spherical harmonics. Evolution (N. Y). 63: 1003–1016.



    """

    def __init__(self, n_harmonics=20, reflect=False, metric="", impute=False):
        # self.dtype = dtype
        self.n_harmonics = n_harmonics

    def fit_transform(self, X, t=None):
        """Fit the model with X.

        Parameters
        ------------------------
        X: list of array-like
                Coordinate values of n_samples. The i-th array-like whose shape (n_coords_i, 2) represents 2D coordinate values of the i-th sample .

        t: list of array-like, optional
                Parameters indicating the position on the outline of n_samples. The i-th ndarray whose shape (n_coords_i, ) corresponds to each coordinate value in the i-th element of X. If `t=None`, then t is calculated based on the coordinate values with the linear interpolation.

        Returns
        ------------------------
        spharm_coef: array-like of shape (n_samples, (1+2*n_harmonics)*n_dim)
            Returns the array-like of coefficients.
        """

        spharm_coef = None

        return spharm_coef

    def _fit_transform_single(self, X, t=None):
        """Fit the model with a signle outline.

        Parameters
        ----------
        X: array-like of shape (n_coords, n_dim)
                Coordinate values of an outline in n_dim (2 or 3).

        t: array-like of shape (n_coords, ), optional
                A parameter indicating the position on the outline.
                If `t=None`, then t is calculated based on the coordinate values with the linear interpolation.

        Returns
        ------------------------
        spharm_coef: list of coeffients
            Returns the coefficients of Fourier series.

        ToDo
        ------------------------
        * EHN: 3D outline
        """

        spharm_coef = None

        return spharm_coef

    # def transform(self, X):

    #     return X_transformed

    def _inverse_transform_single(self, lmax, coef_list, theta_range, phi_range):
        """SPHARM

        Parameters
        ------------------------
        lmax: int
            Degree of SPHARM to use how far
        coef_list: list
            SPHARM coefficients
            coef_naive[l,m] corresponding to the coefficient of degree l and order m
        theta_range: array_like
        phi_range: array_like


        Returns
        ------------------------
        x, y, z: tuple of array_like
            Coordinate values of SPHARM.

        """
        x = 0
        y = 0
        z = 0
        # coef_list = cvt_spharm_coef_from_SlicerSALT_to_list(coef_naive)
        for l in range(lmax + 1):
            m, theta, phi = np.meshgrid(np.arange(-l, l + 1, 1), theta_range, phi_range)
            coef = coef_list[l]
            x = x + np.sum(
                sp.special.sph_harm(m, l, theta, phi) * coef[:, 0].reshape((-1, 1)),
                axis=1,
            )
            y = y + np.sum(
                sp.special.sph_harm(m, l, theta, phi) * coef[:, 1].reshape((-1, 1)),
                axis=1,
            )
            z = z + np.sum(
                sp.special.sph_harm(m, l, theta, phi) * coef[:, 2].reshape((-1, 1)),
                axis=1,
            )

        return np.real(x), np.real(y), np.real(z)


class PCContribDisplay:
    """Shape variation visualization along PC axes.


    Parameters
    ------------------------
    pca:

    n_PCs:

    sd_values:

    lmax:

    theta:

    standardized:


    dpi:

    morph_color:


    morph_alpha:




    Attributes
    ------------------------
    XXX : matplotlib Artist
        .
    ax_ : matplotlib Axes
        Axes with shape variation.
    figure_ : matplotlib Figure
        Figure containing the shape variation.


    """

    def __init__(self):
        pass

    def plot(
        self,
        *,
        include_values=True,
        cmap="viridis",
        xticks_rotation="horizontal",
        values_format=None,
        ax=None,
        colorbar=True,
    ):

        """Plot visualization.
        Parameters
        ------------------------
        include_values : bool, default=True
            Includes values in confusion matrix.
        cmap : str or matplotlib Colormap, default='viridis'
            Colormap recognized by matplotlib.
        xticks_rotation : {'vertical', 'horizontal'} or float, \
                         default='horizontal'
            Rotation of xtick labels.
        values_format : str, default=None
            Format specification for values in confusion matrix. If `None`,
            the format specification is 'd' or '.2g' whichever is shorter.
        ax : matplotlib axes, default=None
            Axes object to plot on. If `None`, a new figure and axes is
            created.
        colorbar : bool, default=True
            Whether or not to add a colorbar to the plot.
        Returns
        ------------------------
        display : :class:`~ktch.outline.PCContribDisplay`
        """

        import matplotlib.pyplot as plt

        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.figure

        self.figure_ = fig
        self.ax_ = ax

        return self

    @classmethod
    def from_estimator(self):
        """Create a"""
        pass



"""
The :mod:`katatch.outline` module implements outline-based morphometrics.
"""

# Copyright 2020 Koji Noshita
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# from ._polar_Fourier_analysis import PFA
from ._elliptic_Fourier_analysis import EllipticFourierAnalysis
from ._spherical_harmonic_analysis import SphericalHarmonicAnalysis, PCContribDisplay

# from ._wavelet_analysis import WaveletAnalysis

# __all__ = ['PFA', 'EllipticFourierAnalysis', 'WaveletAnalysis']
__all__ = ["EllipticFourierAnalysis", "SphericalHarmonicAnalysis", "PCContribDisplay"]



import pytest
import numpy as np

from sklearn.datasets import load_iris
from sklearn.utils.testing import assert_array_equal
from sklearn.utils.testing import assert_allclose

from skltemplate import TemplateEstimator
from skltemplate import TemplateTransformer
from skltemplate import TemplateClassifier


@pytest.fixture
def data():
    return load_iris(return_X_y=True)

def test_template_estimator(data):
    est = TemplateEstimator()
    assert est.demo_param == 'demo_param'

    est.fit(*data)
    assert hasattr(est, 'is_fitted_')

    X = data[0]
    y_pred = est.predict(X)
    assert_array_equal(y_pred, np.ones(X.shape[0], dtype=np.int64))


def test_template_transformer_error(data):
    X, y = data
    trans = TemplateTransformer()
    trans.fit(X)
    with pytest.raises(ValueError, match="Shape of input is different"):
        X_diff_size = np.ones((10, X.shape[1] + 1))
        trans.transform(X_diff_size)


def test_template_transformer(data):
    X, y = data
    trans = TemplateTransformer()
    assert trans.demo_param == 'demo'

    trans.fit(X)
    assert trans.n_features_ == X.shape[1]

    X_trans = trans.transform(X)
    assert_allclose(X_trans, np.sqrt(X))

    X_trans = trans.fit_transform(X)
    assert_allclose(X_trans, np.sqrt(X))


def test_template_classifier(data):
    X, y = data
    clf = TemplateClassifier()
    assert clf.demo_param == 'demo'

    clf.fit(X, y)
    assert hasattr(clf, 'classes_')
    assert hasattr(clf, 'X_')
    assert hasattr(clf, 'y_')

    y_pred = clf.predict(X)
    assert y_pred.shape == (X.shape[0],)



"""
The :mod:`ktch.datasets` module implements utility functions
to load morphometric datasets.
"""

from ._base import (
    load_landmark_mosquito_wings,
    load_outline_mosquito_wings,
    load_outline_bottles,
    load_coefficient_bottles,
)

__all__ = [
    "load_landmark_mosquito_wings",
    "load_outline_mosquito_wings",
    "load_outline_bottles",
    "load_coefficient_bottles",
]



"""
This is a module to be used as a reference for building other modules
"""

# Copyright 2020 Koji Noshita
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin, TransformerMixin
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from sklearn.utils.multiclass import unique_labels
from sklearn.metrics import euclidean_distances


class TemplateEstimator(BaseEstimator):
    """ A template estimator to be used as a reference implementation.

    For more information regarding how to build your own estimator, read more
    in the :ref:`User Guide <user_guide>`.

    Parameters
    ----------
    demo_param : str, default='demo_param'
        A parameter used for demonstation of how to pass and store paramters.

    Examples
    --------
    >>> from skltemplate import TemplateEstimator
    >>> import numpy as np
    >>> X = np.arange(100).reshape(100, 1)
    >>> y = np.zeros((100, ))
    >>> estimator = TemplateEstimator()
    >>> estimator.fit(X, y)
    TemplateEstimator(demo_param='demo_param')
    """
    def __init__(self, demo_param='demo_param'):
        self.demo_param = demo_param

    def fit(self, X, y):
        """A reference implementation of a fitting function.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            The training input samples.
        y : array-like, shape (n_samples,) or (n_samples, n_outputs)
            The target values (class labels in classification, real numbers in
            regression).

        Returns
        -------
        self : object
            Returns self.
        """
        X, y = check_X_y(X, y, accept_sparse=True)
        self.is_fitted_ = True
        # `fit` should always return `self`
        return self

    def predict(self, X):
        """ A reference implementation of a predicting function.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            The training input samples.

        Returns
        -------
        y : ndarray, shape (n_samples,)
            Returns an array of ones.
        """
        X = check_array(X, accept_sparse=True)
        check_is_fitted(self, 'is_fitted_')
        return np.ones(X.shape[0], dtype=np.int64)


class TemplateClassifier(ClassifierMixin, BaseEstimator):
    """ An example classifier which implements a 1-NN algorithm.

    For more information regarding how to build your own classifier, read more
    in the :ref:`User Guide <user_guide>`.

    Parameters
    ----------
    demo_param : str, default='demo'
        A parameter used for demonstation of how to pass and store paramters.

    Attributes
    ----------
    X_ : ndarray, shape (n_samples, n_features)
        The input passed during :meth:`fit`.
    y_ : ndarray, shape (n_samples,)
        The labels passed during :meth:`fit`.
    classes_ : ndarray, shape (n_classes,)
        The classes seen at :meth:`fit`.
    """
    def __init__(self, demo_param='demo'):
        self.demo_param = demo_param

    def fit(self, X, y):
        """A reference implementation of a fitting function for a classifier.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The training input samples.
        y : array-like, shape (n_samples,)
            The target values. An array of int.

        Returns
        -------
        self : object
            Returns self.
        """
        # Check that X and y have correct shape
        X, y = check_X_y(X, y)
        # Store the classes seen during fit
        self.classes_ = unique_labels(y)

        self.X_ = X
        self.y_ = y
        # Return the classifier
        return self

    def predict(self, X):
        """ A reference implementation of a prediction for a classifier.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        y : ndarray, shape (n_samples,)
            The label for each sample is the label of the closest sample
            seen during fit.
        """
        # Check is fit had been called
        check_is_fitted(self, ['X_', 'y_'])

        # Input validation
        X = check_array(X)

        closest = np.argmin(euclidean_distances(X, self.X_), axis=1)
        return self.y_[closest]


class TemplateTransformer(TransformerMixin, BaseEstimator):
    """ An example transformer that returns the element-wise square root.

    For more information regarding how to build your own transformer, read more
    in the :ref:`User Guide <user_guide>`.

    Parameters
    ----------
    demo_param : str, default='demo'
        A parameter used for demonstation of how to pass and store paramters.

    Attributes
    ----------
    n_features_ : int
        The number of features of the data passed to :meth:`fit`.
    """
    def __init__(self, demo_param='demo'):
        self.demo_param = demo_param

    def fit(self, X, y=None):
        """A reference implementation of a fitting function for a transformer.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            The training input samples.
        y : None
            There is no need of a target in a transformer, yet the pipeline API
            requires this parameter.

        Returns
        -------
        self : object
            Returns self.
        """
        X = check_array(X, accept_sparse=True)

        self.n_features_ = X.shape[1]

        # Return the transformer
        return self

    def transform(self, X):
        """ A reference implementation of a transform function.

        Parameters
        ----------
        X : {array-like, sparse-matrix}, shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        X_transformed : array, shape (n_samples, n_features)
            The array containing the element-wise square roots of the values
            in ``X``.
        """
        # Check is fit had been called
        check_is_fitted(self, 'n_features_')

        # Input validation
        X = check_array(X, accept_sparse=True)

        # Check that the input is of the same shape as the one passed
        # during fit.
        if X.shape[1] != self.n_features_:
            raise ValueError('Shape of input is different from what was seen'
                             'in `fit`')
        return np.sqrt(X)



class PCContribDisplay:
    """"""

    def __init__(self, pca, *, display_labels=None) -> None:
        self.pca = pca
        self.display_labels = display_labels

    def plot(
        self, *, n_PCs=(1, 2, 3), sd_values=(-2, -1, 0, 1, 2), , ax=None, im_kw=None
    ):

        return self

    @classmethod
    def from_estimators(cls, estimator, X, y, *, labels=None):
        return cls.from_predictions()

    @classmethod
    def from_predictions(cls, y_true, y_pred, *, labels=None):
        disp = cls(pca=pca)
        return disp.plot()



"""Procrustes Analysis"""

# Copyright 2020 Koji Noshita
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import scipy as sp

from sklearn.base import BaseEstimator, TransformerMixin


class GeneralizedProcrustesAnalysis(TransformerMixin, BaseEstimator):
    """
    Generalized Procrustes Analysis (GPA)

    Notes
    ------------

    test

    """

    def __init__(self, tol=10 ^ -10):
        self.tol = tol

    def fit(self, X):
        Gamma = 0

        return self

    def transform(self, X):
        X_ = X
        X_ = self._translate(X_)
        X_ = self._scale(X_)
        mu = np.sum(X_, axis=0) / len(X_)

        d_Procrustes_dist = 10 * self.tol
        while d_Procrustes_dist < self.tol:
            results = [sp.spatial.procrustes(mu, x) for x in X]
            X_ = np.array([result[1] for result in results])
            d_Procrustes_dist = np.sum(np.array([result[2] for result in results]))
            mu = np.sum(X_, axis=0) / len(X_)

        return X_

    def fit_transform(self, X):
        return self.transform(X)

    def _translate(self, X):
        X_translated = np.array([x - np.mean(x, axis=0) for x in X])
        return X_translated

    def _scale(self, X):
        X_scaled = np.array(
            [x / np.sqrt(np.trace(np.dot(x, x.transpose()))) for x in X]
        )
        return X_scaled


class LandmarkImputer(TransformerMixin, BaseEstimator):
    def __init__(self, missing_values=np.nan):
        pass



"""Elliptic Fourier Analysis"""

# Copyright 2020 Koji Noshita
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from operator import index
import numpy as np
import scipy as sp
from scipy.spatial.transform import Rotation as R
import pandas as pd

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.base import ClassNamePrefixFeaturesOutMixin


class EllipticFourierAnalysis(TransformerMixin, BaseEstimator):
    r"""
    Elliptic Fourier Analysis (EFA) 

    Parameters
    ------------
    n_harmonics: int, default=20
        harmonics

    reflect: bool, default=False
        reflect

    metric: str
        metric

    impute: bool, False
        impute

    Notes
    ------------
    :cite:`EFA-KUHL:1982bq`

    [Kuhl_Giardina_1982]_

    .. math::
        \begin{align}
            x(l) &=
            \frac{a_0}{2} + \sum_{i=1}^{n}
            \left[ a_i \cos\left(\frac{2\pi i t}{T}\right)
            + b_i \sin\left(\frac{2\pi i t}{T}\right) \right]\\
            y(l) &=
            \frac{c_0}{2} + \sum_{i=1}^{n}
            \left[ c_i \cos\left(\frac{2\pi i t}{T}\right)
            + d_i \sin\left(\frac{2\pi i t}{T}\right) \right]\\
        \end{align}

    References
    ------------
    .. [Kuhl_Giardina_1982] Kuhl, F.P., Giardina, C.R. (1982) Elliptic Fourier features of a closed contour. Comput. Graph. Image Process. 18: 236–258.


    """

    def __init__(self, n_harmonics=20, reflect=False, metric="", impute=False):
        # self.dtype = dtype
        self.n_harmonics = n_harmonics
        self.reflect = reflect
        self.metric = metric
        self.impute = impute

    def transform(self, X, t=None):
        return self.fit_transform(X, t)

    def fit_transform(self, X, t=None, as_frame=False):
        """

        Fit the model with X.

        Parameters
        ------------
        X: {list of array-like, array-like} of shape (n_samples, n_coords, 2)
            Coordinate values of n_samples.
            The i-th array-like of shape (n_coords_i, 2) represents
            2D coordinate values of the i-th sample.

        t: array-like of shape (n_samples, n_coords), optional
            Parameters indicating the position on the outline of n_samples.
            The i-th ndarray of shape (n_coords_i, ) corresponds to
            each coordinate value in the i-th element of X.
            If `t=None`, then t is calculated based on
            the coordinate values with the linear interpolation.

        Returns
        ------------
        X_transformed: array-like of shape (n_samples, (1+2*n_harmonics)*n_dim)
            Returns the array-like of coefficients.

        """

        n_harmonics = self.n_harmonics

        if t is None:
            t_ = [None for i in range(len(X))]

        if len(t_) != len(X):
            raise ValueError("t must have a same length of X ")

        if as_frame:
            X_transformed = pd.concat(
                [
                    self._fit_transform_single(X[i], t=t_[i], as_frame=True)
                    for i in range(len(X))
                ],
                axis=0,
            )
            X_transformed["specimen_id"] = [
                i for i in range(len(X)) for j in range(n_harmonics + 1)
            ]
            X_transformed = X_transformed.reset_index().set_index(
                ["specimen_id", "harmonics"]
            )
        else:
            if isinstance(X, pd.DataFrame):
                X_ = [x[0] for x in X.to_numpy()]
            else:
                X_ = X

            X_transformed = np.stack(
                [
                    self._fit_transform_single(np.array(X_[i]), t_[i], as_frame=False)
                    for i in range(len(X))
                ]
            )

        return X_transformed

    def _fit_transform_single(
        self,
        X,
        t=None,
        norm=True,
        duplicated_points="infinitesimal",
        as_frame=False,
    ):
        """Fit the model with a signle outline.

        Parameters
        ----------
        X: array-like of shape (n_coords, 2)
                Coordinate values of an 2D outline.

        t: array-like of shape (n_coords+1, ), optional
                A parameter indicating the position on the outline.
                Both t[0] and t[n_coords] corresponds to X[0].
                If `t=None`, then t is calculated based on
                the coordinate values with the linear interpolation.

        Returns
        -------
        X_transformed: list of coeffients
            Returns the coefficients of Fourier series.

        ToDo
        -------
        * EHN: 3D outline
        """

        X_arr = np.array(X)
        n_harmonics = self.n_harmonics

        dx = np.append(
            X_arr[0, 0] - X_arr[-1, 0],
            X_arr[1:, 0] - X_arr[:-1, 0],
        )
        dy = np.append(
            X_arr[0, 1] - X_arr[-1, 1],
            X_arr[1:, 1] - X_arr[:-1, 1],
        )

        if t is None:
            dt = np.sqrt(dx**2 + dy**2)
            tp = np.append(0, np.cumsum(dt))
            T = np.sum(dt)
        else:
            # TODO: add test
            dt = t[1:] - t[:-1]
            tp = t
            T = t[-1]

        if duplicated_points == "infinitesimal":
            dt[dt < 10**-10] = 10**-10
        elif duplicated_points == "deletion":
            idx_duplicated_points = np.where(dt == 0)[0]
            if len(idx_duplicated_points) > 0:
                dx = np.delete(dx, idx_duplicated_points)
                dy = np.delete(dy, idx_duplicated_points)
                dt = np.delete(dt, idx_duplicated_points)
                tp = np.delete(
                    tp,
                    (np.array(idx_duplicated_points) + 1).tolist(),
                )
                X_arr = np.delete(X_arr, idx_duplicated_points, 0)
        else:
            raise ValueError(
                "'duplicated_points' must be 'infinitesimal' or 'deletion'"
            )

        if len(tp) != len(X_arr) + 1:
            raise ValueError(
                "len(t) must have a same len(X) + 1), len(t): "
                + str(len(tp))
                + ", len(X)+1: "
                + str(len(X_arr))
            )

        a0 = 2 / T * np.sum(X_arr[:, 0] * dt)
        c0 = 2 / T * np.sum(X_arr[:, 1] * dt)

        an = [
            (T / (2 * (np.pi**2) * (n**2)))
            * np.sum(
                (dx / dt)
                * (
                    np.cos(2 * np.pi * n * tp[1:] / T)
                    - np.cos(2 * np.pi * n * tp[:-1] / T)
                )
            )
            for n in range(1, n_harmonics + 1, 1)
        ]
        bn = [
            (T / (2 * (np.pi**2) * (n**2)))
            * np.sum(
                (dx / dt)
                * (
                    np.sin(2 * np.pi * n * tp[1:] / T)
                    - np.sin(2 * np.pi * n * tp[:-1] / T)
                )
            )
            for n in range(1, n_harmonics + 1, 1)
        ]
        cn = [
            (T / (2 * (np.pi**2) * (n**2)))
            * np.sum(
                (dy / dt)
                * (
                    np.cos(2 * np.pi * n * tp[1:] / T)
                    - np.cos(2 * np.pi * n * tp[:-1] / T)
                )
            )
            for n in range(1, n_harmonics + 1, 1)
        ]
        dn = [
            (T / (2 * (np.pi**2) * (n**2)))
            * np.sum(
                (dy / dt)
                * (
                    np.sin(2 * np.pi * n * tp[1:] / T)
                    - np.sin(2 * np.pi * n * tp[:-1] / T)
                )
            )
            for n in range(1, n_harmonics + 1, 1)
        ]

        an = [a0] + an
        bn = [0] + bn
        cn = [c0] + cn
        dn = [0] + dn

        if norm:
            an, bn, cn, dn = self._normalize(an, bn, cn, dn)

        if as_frame:
            harmonics = pd.Series([i for i in range(n_harmonics + 1)])
            df_a = pd.DataFrame(an, index=harmonics)
            df_b = pd.DataFrame(bn, index=harmonics)
            df_c = pd.DataFrame(cn, index=harmonics)
            df_d = pd.DataFrame(dn, index=harmonics)
            X_transformed = pd.concat([df_a, df_b, df_c, df_d], axis=1)
            X_transformed.columns = ["an", "bn", "cn", "dn"]
            X_transformed.index.name = "harmonics"
        else:
            # X_transformed = np.stack([an, bn, cn, dn], axis=1)
            X_transformed = np.concatenate([an, bn, cn, dn]).reshape(-1)

        return X_transformed

    def _normalize(self, an, bn, cn, dn):
        a1 = an[1]
        b1 = bn[1]
        c1 = cn[1]
        d1 = dn[1]

        theta = (1 / 2) * np.arctan2(
            2 * (a1 * b1 + c1 * d1), (a1**2 + c1**2 - b1**2 - d1**2)
        )
        if theta < 0:
            theta = theta + np.pi

        a_s = a1 * np.cos(theta) + b1 * np.sin(theta)
        c_s = c1 * np.cos(theta) + d1 * np.sin(theta)
        scale = np.sqrt(a_s**2 + c_s**2)
        psi = np.arctan2(c_s, a_s)
        if psi < 0:
            psi = 2 * np.pi + psi

        coef_norm_list = []
        r_psi = rotaion_matrix_2d(-psi)
        for n in range(1, len(an)):
            r_ntheta = rotaion_matrix_2d(n * theta)
            coef_orig = np.array([[an[n], bn[n]], [cn[n], dn[n]]])
            coef_norm_tmp = (1 / scale) * np.dot(np.dot(r_psi, coef_orig), r_ntheta)
            coef_norm_list.append(coef_norm_tmp.reshape(-1))

        coef_norm = np.stack(coef_norm_list)
        An = np.append(an[0], coef_norm[:, 0])
        Bn = np.append(bn[0], coef_norm[:, 1])
        Cn = np.append(cn[0], coef_norm[:, 2])
        Dn = np.append(dn[0], coef_norm[:, 3])

        return An, Bn, Cn, Dn

    def _inverse_transform_single(
        self,
        X_transformed,
        t_num=100,
        norm=True,
        as_frame=False,
    ):
        coef = X_transformed
        if as_frame:
            an = coef["an"][1:]
            bn = coef["bn"][1:]
            cn = coef["cn"][1:]
            dn = coef["dn"][1:]
        else:
            an, bn, cn, dn = coef.reshape([4, -1])
            an = an[1:]
            bn = bn[1:]
            cn = cn[1:]
            dn = dn[1:]

        n_max = len(an)

        theta = np.linspace(0, 2 * np.pi, t_num)

        cos = np.cos(np.tensordot(np.arange(1, n_max + 1, 1), theta, 0))
        sin = np.sin(np.tensordot(np.arange(1, n_max + 1, 1), theta, 0))

        x = np.dot(an, cos) + np.dot(bn, sin)
        y = np.dot(cn, cos) + np.dot(dn, sin)

        X_coords = np.stack([x, y], 1)

        return X_coords

    def inverse_transform(self, X_transformed, t_num=100, as_frame=False):
        X_list = []
        sp_num = X_transformed.shape[0]

        for i in range(sp_num):

            if as_frame:
                coef = X_transformed.loc[i]
                X = self._inverse_transform_single(coef, as_frame=as_frame)
                df_X = pd.DataFrame(X, columns=["x", "y"])
                df_X["coord_id"] = [coord_id for coord_id in range(len(X))]
                df_X["specimen_id"] = i
                X_list.append(df_X)
            else:
                coef = X_transformed[i]
                X = self._inverse_transform_single(coef, as_frame=as_frame)
                X_list.append(X)

        if as_frame:
            X_coords = pd.concat(X_list)
            X_coords = X_coords.set_index(["specimen_id", "coord_id"])
        else:
            X_coords = X_list

        return X_coords

    def get_feature_names_out(self):
        an = ["a_" + str(i) for i in range(self.n_harmonics + 1)]
        bn = ["b_" + str(i) for i in range(self.n_harmonics + 1)]
        cn = ["c_" + str(i) for i in range(self.n_harmonics + 1)]
        dn = ["d_" + str(i) for i in range(self.n_harmonics + 1)]
        return np.asarray(an + bn + cn + dn, dtype=object)

    @property
    def _n_features_out(self):
        """Number of transformed output features."""
        return (self.n_harmonics + 1) * 4


def rotaion_matrix_2d(theta):
    rot_mat = np.array(
        [[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]]
    )
    return rot_mat



# Copyright 2020 Koji Noshita
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import types

from . import landmark
from . import outline

from ._template import TemplateEstimator, TemplateTransformer, TemplateClassifier
from ._version import __version__


__all__ = [
    "landmark",
    "outline",
    "datasets",
    "TemplateEstimator",
    "TemplateTransformer",
    "TemplateClassifier",
    "__version__",
]



"""Base IO code for small sample datasets"""

# Copyright 2020 Koji Noshita
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from importlib import resources

import pandas as pd
from sklearn.utils import Bunch
from sklearn.datasets._base import load_descr


def load_landmark_mosquito_wings(*, as_frame=True):
    """Load and return the mosquito wing landmark dataset
    (landmark-based morphometrics).

    ========================   ============
    Specimens                      127
    Landmarks per specimen          18
    Landmark dimensionality          2
    Features                      real
    ========================   ============

    Parameters
    ----------
    as_frame : bool, default=False
        If True, the data is a pandas DataFrame including columns with
        appropriate dtypes (numeric). The target is
        a pandas DataFrame or Series depending on the number of target columns.
        If `return_X_y` is True, then (`data`, `target`) will be pandas
        DataFrames or Series as described below.

    Returns
    ----------
    data : :class:`~sklearn.utils.Bunch`
        Dictionary-like object, with the following attributes.
        coords : {ndarray, dataframe} of shape (150, 4)
            The data matrix. If `as_frame=True`, `data` will be a pandas
            DataFrame.
        meta : {dataframe, list} of shape ()
        DESCR: str
            The full description of the dataset.
        filename: str
            The path to the location of the data.

    """
    data_module = "ktch.datasets.data"
    data_file_name = "data_landmark_mosquito_wings.csv"
    metadata_file_name = "meta_landmark_mosquito_wings.csv"
    descr_module = "ktch.datasets.descr"
    descr_file_name = "data_landmark_mosquito_wings.rst"

    coords = pd.read_csv(
        resources.open_text(data_module, data_file_name), index_col=[0, 1]
    )
    meta = pd.read_csv(
        resources.open_text(data_module, metadata_file_name), index_col=[0]
    )
    fdescr = load_descr(
        descr_module=descr_module,
        descr_file_name=descr_file_name,
    )

    if not as_frame:
        coords = coords.to_numpy()
        meta = meta.to_dict()

    return Bunch(
        coords=coords,
        meta=meta,
        DESCR=fdescr,
        filename=data_file_name,
    )


def load_outline_mosquito_wings(*, as_frame=True):
    """Load and return the mosquito wing outline dataset
    (outline-based morphometrics).

    ========================   ===========
    Specimens                      126
    Landmarks per specimen         100
    Landmark dimensionality          2
    Features                      real
    ========================   ===========

    Parameters
    ----------
    as_frame : bool, default=False
        If True, the data is a pandas DataFrame including columns with
        appropriate dtypes (numeric). The target is
        a pandas DataFrame or Series depending on the number of target columns.
        If `return_X_y` is True, then (`data`, `target`) will be pandas
        DataFrames or Series as described below.

    Returns
    ----------
    data : :class:`~sklearn.utils.Bunch`
        Dictionary-like object, with the following attributes.
        coords : {ndarray, dataframe} of shape (150, 4)
            The data matrix. If `as_frame=True`, `data` will be a pandas
            DataFrame.
        meta : {dataframe, list} of shape ()
        DESCR: str
            The full description of the dataset.
        filename: str
            The path to the location of the data.

    """
    data_module = "ktch.datasets.data"
    data_file_name = "data_outline_mosquito_wings.csv"
    metadata_file_name = "meta_outline_mosquito_wings.csv"
    descr_module = "ktch.datasets.descr"
    descr_file_name = "data_outline_mosquito_wings.rst"

    coords = pd.read_csv(
        resources.open_text(data_module, data_file_name), index_col=[0, 1]
    )
    meta = pd.read_csv(
        resources.open_text(data_module, metadata_file_name), index_col=[0]
    )
    fdescr = load_descr(
        descr_module=descr_module,
        descr_file_name=descr_file_name,
    )

    if not as_frame:
        coords = coords.to_numpy()
        meta = meta.to_dict()

    return Bunch(
        coords=coords,
        meta=meta,
        DESCR=fdescr,
        filename=data_file_name,
    )


def load_outline_bottles(*, as_frame=True):
    """Load and return the Momocs bottle outline dataset
    (outline-based morphometrics).

    =================   ==============
    Specimens                       40
    dimensionality                   4
    Features                      real
    =================   ==============

    Parameters
    ----------
    as_frame : bool, default=False
        If True, the data is a pandas DataFrame including columns with
        appropriate dtypes (numeric). The target is
        a pandas DataFrame or Series depending on the number of target columns.

    Returns
    -------
    data : :class:`~sklearn.utils.Bunch`
        Dictionary-like object, with the following attributes.
        coords : {list of ndarray, dataframe} of shape (40, n_coords, 2)
            The data matrix. If `as_frame=True`, `coords` will be a pandas
            DataFrame.
        meta : {dataframe, list} of shape ()
        DESCR: str
            The full description of the dataset.
        filename: str
            The path to the location of the data.

    """
    data_module = "ktch.datasets.data"
    data_file_name = "data_outline_bottles.csv"
    metadata_file_name = "meta_outline_bottles.csv"
    descr_module = "ktch.datasets.descr"
    descr_file_name = "data_outline_bottles.rst"

    coords = pd.read_csv(
        resources.open_text(data_module, data_file_name), index_col=[0, 1]
    )
    meta = pd.read_csv(
        resources.open_text(data_module, metadata_file_name), index_col=[0]
    )
    fdescr = load_descr(
        descr_module=descr_module,
        descr_file_name=descr_file_name,
    )

    if not as_frame:
        coords = coords.to_numpy()
        meta = meta.to_dict()

    return Bunch(
        coords=coords,
        meta=meta,
        DESCR=fdescr,
        filename=data_file_name,
    )


def load_coefficient_bottles(*, as_frame=True, norm=True):
    """Load and return the coefficients of  Momocs bottle outline datasets
    for testing the EFA

    Parameters
    ----------
    as_frame (bool, optional):
        If True, the data is a pandas DataFrame
        including columns with appropriate dtypes (numeric).
        The target is a pandas DataFrame or Series
        depending on the number of target columns.
    norm: bool, optional

    Returns
    -------
    data : :class:`~sklearn.utils.Bunch`
        Dictionary-like object, with the following attributes.
        coef : {list of ndarray, dataframe} of shape (40, 6, 4)
            The data matrix. If `as_frame=True`, `coords` will be a pandas
            DataFrame.
        meta : {dataframe, list} of shape ()
        DESCR: str
            The full description of the dataset.
        filename: str
            The path to the location of the data.
    """

    data_module = "ktch.datasets.data"
    if norm:
        data_file_name = "test_coef_nharm_6_norm_true_outline_bottles.csv"
    else:
        data_file_name = "test_coef_nharm_6_norm_false_outline_bottles.csv"
    metadata_file_name = "meta_outline_bottles.csv"
    descr_module = "ktch.datasets.descr"
    descr_file_name = "data_outline_bottles.rst"

    coef = pd.read_csv(
        resources.open_text(data_module, data_file_name), index_col=[0, 1]
    )
    meta = pd.read_csv(
        resources.open_text(data_module, metadata_file_name), index_col=[0]
    )
    fdescr = load_descr(
        descr_module=descr_module,
        descr_file_name=descr_file_name,
    )

    if not as_frame:
        coef = coef.to_numpy()
        meta = meta.to_dict()

    return Bunch(
        coef=coef,
        meta=meta,
        DESCR=fdescr,
        filename=data_file_name,
    )



from numpy.testing import assert_array_almost_equal
import pytest

from ktch.datasets import load_outline_bottles, load_coefficient_bottles
from ktch.outline import EllipticFourierAnalysis

bottles = load_outline_bottles()
bottles_coef = load_coefficient_bottles(norm=False)

# @pytest.mark.parameterize("norm", [False, True])
def test_efa():
    n_harmonics = 6
    X = [bottles.coords.loc[i].to_numpy() for i in range(1, 41, 1)]
    efa = EllipticFourierAnalysis(n_harmonics=n_harmonics)

    X_transformed = efa.fit_transform(X)
    coef = [X_transformed.loc[i].loc[1:].to_numpy() for i in range(40)]
    coef_val = [bottles_coef.coef.loc[i].to_numpy() for i in range(1, 41, 1)]

    assert_array_almost_equal(coef, coef_val)
